// TreeDraw1.cpp: implementation of the CTreeDraw class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TreeDraw.h"
#include "resource.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#define IDENT 12
#define DELTA 10
#define ICON_SIZE 16
#define DY 20
#define MIN_DELTA 20
#define SCROLL_STEP 200


CTreeDraw::CTreeDraw()
{
	TreeRoot=LightNode=0;
	m_font.CreateFont(14,0,0,0,0,0,0,0,0,0,0,0,0,_T("Courier"));
}


CTreeDraw::~CTreeDraw()
{
	Clear();
}


//    
void CTreeDraw::CalcSize(CDC *pDC,tree_node_st *root,int x)
{
	char *end,*first;
	if(x%DELTA) x+=DELTA-x%DELTA; 
	while(root) {
		if(!root->len) {
			root->x=x;
			first=root->real_node->first;
			end=root->real_node->last; if(!end) end=end_text;
			root->len=pDC->GetTextExtent(first,end-first).cx+5;
			if(root->real_node->node) root->len+=ICON_SIZE;
			else if(root->len>SCROLL_STEP) root->len+=ICON_SIZE;
		}
		root->scroll=root->len;
		if(!root->child && root->scroll>SCROLL_STEP) root->scroll=SCROLL_STEP;
		CalcSize(pDC,root->child,x+root->len+IDENT);
		root=root->next;
	}
}

//  x    
void CTreeDraw::CalcArray(tree_node_st *node)
{
	if(!node) return;
	m_map.insert(std::pair<int,int>(node->x,0));
	while(node) {
		if(node->open) CalcArray(node->child);
		node=node->next;
	}
}


//        
// (    )
int CTreeDraw::CorrectMap(tree_node_st *node)
{
	while(node->next) node=node->next;
	if(node->child && node->open) return CorrectMap(node->child);
	return node->x+node->scroll;
}

//       
// (    )
void CTreeDraw::CorrectMap(tree_node_st *node,int x1,int corr)
{
	int x2;	std::map<int,int>::iterator it;
	if(!node) return;
	x2=CorrectMap(node);
	for(it=m_map.begin(); it!=m_map.end(); it++) {
		if(it->first>x2) break;
		if(it->first>=x1) it->second+=corr;
	}
}



//    
CPoint CTreeDraw::CalcSpace(tree_node_st *node)
{
	CPoint pt;
	int xx; int min=-1,max;
	std::map<int,int>::iterator it;
	if(!node) return 0;
	while(node) {
		node->corr=max=0;
		xx=node->x+node->scroll+IDENT/2;
		for(it=m_map.begin(); it!=m_map.end(); it++) {
			if(it->first>xx) break;
			if(it->first>=node->x && it->second>max) max=it->second;
		}
		node->y=max+DY;
		if(node->open) {
			pt=CalcSpace(node->child);
			if(pt.x-node->y>MIN_DELTA) node->y=pt.x-MIN_DELTA/2;
		}
		if(min<0) min=node->y;
		for(it=m_map.begin(); it!=m_map.end(); it++) {
			if(it->first>xx) break;
			if(it->first>=node->x) it->second=node->y;
		}
		if(node->open && node->y-pt.y>MIN_DELTA) 
			node->corr=node->y-pt.y-MIN_DELTA-MIN_DELTA/2,CorrectMap(node->child,xx,node->corr);
		node=node->next;
	}
	return CPoint(min,max);
}


CSize CTreeDraw::DrawNode(CDC *pDC,CPoint pt,tree_node_st *node,int corr)
{
	int n,id,y; CRect rect; char *first,*end;
	CSize size1,size2;
	size1.cx=size1.cy=0;
	while(node) {
		id=0;
		n=node->x+node->scroll;
		if(node->child) id=(node==LightNode)?(node->open?IDI_L_MINUS:IDI_L_PLUS):(node->open?IDI_MINUS:IDI_PLUS);
		else if(node->len>SCROLL_STEP) id=node==LightNode?(light_offset?IDI_L1_MOVE:IDI_L2_MOVE):IDI_MOVE,n=node->x+node->scroll;
		y=node->y+corr;
		pDC->MoveTo(pt);
		pDC->LineTo(node->x,y);
		pDC->LineTo(n,y);
		rect.SetRect(node->x+3,y-14,n,y-1);
		if(id) rect.right-=ICON_SIZE+2;
		first=node->real_node->first;
		end=node->real_node->last; if(!end) end=end_text;
		pDC->DrawText(first,end-first,rect,DT_LEFT|DT_END_ELLIPSIS);
		if(node->open) {
			size2=DrawNode(pDC,CPoint(n,y),node->child,corr+node->corr);
			if(size2.cx>size1.cx) size1.cx=size2.cx;
			if(size2.cy>size1.cy) size1.cy=size2.cy;
		}
		if(id)DrawIconEx(pDC->m_hDC,n-ICON_SIZE,y-9,AfxGetApp()->LoadIcon(id),16,16,0,0,DI_NORMAL);
		if(n>size1.cx) size1.cx=n;
		if(y>size1.cy) size1.cy=y;
		node=node->next;
	}
	return size1;
}



CSize CTreeDraw::Draw(CDC* pDC)
{
	CSize size;
	CFont *p_font=pDC->SelectObject(&m_font);
	pDC->SetBkMode(TRANSPARENT);
	size=DrawNode(pDC,CPoint(0,0),TreeRoot,0);
	pDC->SelectObject(p_font);
	return size;
}


void CTreeDraw::Mouse(tree_node_st *node,CPoint pt,int corr)
{
	int x,y; CRect rect;
	while(node) {
		if(node->child || node->len>SCROLL_STEP) {
			x=node->x+node->scroll;
			y=node->y+corr;
			rect.SetRect(x-ICON_SIZE,y-ICON_SIZE/2,x,y+ICON_SIZE/2);
			if(rect.PtInRect(pt)) {
				LightNode=node;
				if(!node->child) {light_offset=(pt.x-x)<-ICON_SIZE/2;
					if(light_offset && node->scroll==SCROLL_STEP) LightNode=0;
					if(!light_offset && node->scroll==node->len) LightNode=0;
				}
			}
		}
		if(node->open) Mouse(node->child,pt,corr+node->corr);
		node=node->next;
	}
}

bool CTreeDraw::Mouse(CPoint point)
{
	bool off=light_offset;
	tree_node_st *old=LightNode;
	LightNode=0;
	Mouse(TreeRoot,point,0);
	return old!=LightNode || off!=light_offset;
}

void CTreeDraw::Click(CDC *pDC)
{
	if(!LightNode) return; 
	CFont *p_font=pDC->SelectObject(&m_font);
	if(!LightNode->calc && LightNode->child) {
		LightNode->child=ExpandNode(LightNode->child,LightNode->child->real_node,2);
		CalcSize(pDC,LightNode,LightNode->x);
		LightNode->calc=true;
	}
	if(LightNode->child) LightNode->open=!LightNode->open;
	else {
		if(light_offset) {
			if(LightNode->scroll%SCROLL_STEP) LightNode->scroll-=LightNode->scroll%SCROLL_STEP;
			else LightNode->scroll-=SCROLL_STEP;
			if(LightNode->scroll<SCROLL_STEP) LightNode->scroll=SCROLL_STEP;		
		}else{
			LightNode->scroll+=SCROLL_STEP; if(LightNode->scroll>LightNode->len) LightNode->scroll=LightNode->len;
		}
		LightNode=0;
	}
	CalcArray(TreeRoot);
	CalcSpace(TreeRoot);
	m_map.clear();
	pDC->SelectObject(p_font);
}


void CTreeDraw::Clear(tree_node_st *node)
{
	if(!node) return;
	Clear(node->next);
	Clear(node->child);
	delete node;
}

void CTreeDraw::Clear()
{
	Clear(TreeRoot);
	TreeRoot=0;
}

tree_node_st *CTreeDraw::ExpandNode(tree_node_st *node,link_st *link,int depth)
{
	if(!link) return 0;
	if(!depth) return node;
	if(link && !node) {
		node=new tree_node_st;
		node->len=0;
		node->next=node->child=0;
		node->open=node->calc=false;
		node->real_node=link;
	}
	node->next=ExpandNode(node->next,link->next,depth);
	if(link->node) node->child=ExpandNode(node->child,link->node->links,depth-1);
	return node;

/*
	tree_node_st *ret=node;
	while(node) {
		if(link->next && !node->next) node->next=ExpandNode(node->next,link->next,depth);
		if(link->node) node->child=ExpandNode(node->child,link->node->links,depth-1);
		node=node->next;
		link=link->next;
	}
	return ret;
*/
}


void CTreeDraw::SetRoot(CDC *pDC,link_st *link,char *end)
{
	Clear();
	end_text=end;
	TreeRoot=ExpandNode(0,link,2);
	CalcSize(pDC,TreeRoot,3);
	CalcArray(TreeRoot);
	CalcSpace(TreeRoot);
	m_map.clear();
}
